; From: "Mitchell Wand" <wand@ccs.neu.edu>
; To: "Gregor Kiczales" <gregor@cs.ubc.ca>
; Sent: Wednesday, May 01, 2002 7:24 PM
; Subject: [Fwd: Mitchell Wand] weaver=traversal.scm


; Here is the code I sent you earlier for calculating which edges to
; traverse.  The main procedure is first-fields which takes two classes,
; c1 and c2, and returns a list of fields of an object of c1 that might
; lead to an object of type c2 (ie, of class <= c2).

; The front end would have to be rejiggered to collect its info from the
; ASB class system rather than from the relations as done here (or
; rather we'd have to attach some advice to elaborate-class! to collect
; the info for first-fields).

; --Mitch 

; ------- start of forwarded message (RFC 934 encapsulation) -------
; MIME-Version: 1.0
; Content-Length: 14116
; Content-Type: text/plain; charset=us-ascii
; Content-Transfer-Encoding: 7bit
; Return-Path: <wand@ccs.neu.edu>
; Delivered-To: wand@alnilam.ccs.neu.edu
; In-Reply-To: <000b01c1da72$d2fa23e0$ce0e678e@cs.ubc.ca>
; References: <000b01c1da72$d2fa23e0$ce0e678e@cs.ubc.ca>
; X-Mailer: VM 6.95 under Emacs 19.34.1
; From: Mitchell Wand <wand@ccs.neu.edu>
; To: "Gregor Kiczales" <gregor@cs.ubc.ca>
; Cc: Karl Lieberherr <lieberherr@ccs.neu.edu>
; Subject: weaver=traversal.scm
; Date: Tue, 2 Apr 2002 16:25:44 -0500

; Karl will be in DC, not me.  Here is some code that I wrote last
; February for figuring out where to explore.  It's based on a setup
; that Ken Anderson did, and my paper with Karl on navigation in object
; graphs.

; I have no idea whether or not it will be adaptable.  I've made only
; the most minor efforts in tree-shaking, but I did it run it just now
; and it worked for me under Chez Scheme.

; - --Mitch 

;;; Mitch's version of Ken's code for composing relations.
;;; Sample run:

; > (load "mitch3.scm")
; > (run-tests)
; test: test1
; (first-fields (quote library) (quote magazine))
; correct outcome: (publications)
; actual outcome:  (publications)
; correct

; test: test2
; (first-fields (quote library) (quote string))
; correct outcome: (name publications)
; actual outcome:  (name publications)
; correct

; test: test3
; (first-fields (quote library) (quote date))
; correct outcome: (publications)
; actual outcome:  (publications)
; correct

; no bugs found
; #t

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; Ken writes:

; Here's my code.
; Class hierarchy is a library system.

; Relations are like
; (subclass Book Published)

; Fields have 4 elements, one for the field name that i carry along.

; (field Library Publications publications)

; The first field is unused, but i thought it would be nice for debugging.
; I think if i do the transitive closure of subclass everything would work with just compose.  But it's simple this way.

; I had trouble adding reflection to the transitive closure so maybe you can suggest an alternative way of doing things.

; Here are some examples:
; > (show (first-fields 'Library 'Magazine))
; (field Library Magazine publications)
; ()
; > (show (first-fields 'Library 'String))
; (field Library String name)
; (field Library String publications)
; ()
; > (show (first-fields 'Library 'Date))
; (field Library Date publications)

;;; class := `(,name ,supers ,fields)
;;; field := `(,class ,name)

(define classes
  `((Object ()
	    ())
    (String (Object) ())
    (Date (Object) ())
    (Named (Object)
	   ((String name)))
    (Author (Named)
	    ())
    (Authors (Object)
	     ())
    (AuthorsCons (Authors)
		 ((Author car)
		  (Authors cdr)))
    (Publisher (Named)
	       ())
    (Authored (Object)
	      ((String title)
	       (Authors authors)
	       (Date date)))
    (Published (Object)
	       ((Publisher publisher)))
    (Book (Published)
	  ((String isbn)))
    (Magazine (Published)
	      ((Articles articles)))
    (Article (Authored)
	     ((Magazine magazine)))
    (Articles (Object) ())
    (ArticlesCons (Articles)
		  ((Article car)
		   (Articles cdr)))
    (Library (Named)
	     ((Publications publications)))
    (Publications (Object)
		  ())
    (PublicationCons (Publications)
		     ((Published car)
		      (Publications cdr)))
    ))

;;; classes is a list of class definitions.
(define (find-class name) (assq name classes))
(define class-name car)
(define class-fields caddr)

;;; For now, just do single inheritance like everyone else.
(define (class-superclass c)
  (if (null? (cadr c)) #f
      (car (cadr c))))

(define tests
  (list
   (equal? (class-name (find-class 'Object)) 'Object)
   (equal? (class-superclass (find-class 'Object)) #f)
   (equal? (class-superclass (find-class 'Author)) 'Object)))

(define field-name cadr)
(define field-type car)

(define (union xs ys)
  (if (null? xs) ys
      (if (member (car xs) ys)
	  (union (cdr xs) ys)
	  (cons (car xs) (union (cdr xs) ys)))))

;; inner loop added for debugging-- not to worry.

(define (remove-duplicates xs) 
  (define (remove-duplicates-internal xs)
    (if (null? xs) '()
	(if (member (car xs) (cdr xs))
	    (remove-duplicates-internal (cdr xs))
	    (cons (car xs) (remove-duplicates-internal (cdr xs))))))
  (remove-duplicates-internal xs))

(set! tests (append tests
    (list (equal? (remove-duplicates '(1 2 3 2 1)) '(3 2 1)))))

(define (grind-class c)
  (let ((s (class-superclass c))
	(fields (map (lambda (f) (grind-field c f)) (class-fields c))))
    `((subclass ,(class-name c) ,s)
      (superclass ,s ,(class-name c))
      ,@ fields)))
    
(define (grind-field c f)
  `(field ,(class-name c) ,(field-type f) ,(field-name f)))

(define relations (apply append (map grind-class classes)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;; Mitch's take on all this: ;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; ok, so you are setting up "relations" as a Prolog-style assertion
;; database.  I can deal with that.

;; but most relations are anonymous! -- so I will represent them as
;; list of ordered 2-lists without duplication.

;;;;;;;;;;;;;;;; a little infrastructure ;;;;;;;;;;;;;;;;

;; Our basic comprehension operator.  map f on all elements of l that
;; satisfy p, ie {f(x) | x \in l & p(x)}.  "filter" is probably a
;; suboptimal name for this.

(define filter
  (lambda (p f l)
    (cond
      ((null? l) '())
      ((p (car l)) (cons (f (car l)) (filter p f (cdr l))))
      (else (filter p f (cdr l))))))

;; take the union of a set of sets
(define (union* ll)
  (if (null? ll) '()
    (union (car ll) (union* (cdr ll)))))


;;;;;;;;;;;;;;;; relation-processing procedures ;;;;;;;;;;;;;;;;

(define (is-in? x y rel)
  (let ((the-pair (list x y)))
    (let loop ((rel rel))
      (cond
        ((null? rel) #f)
        ((equal? (car rel) the-pair) #t)
        (else (loop (cdr rel)))))))

(define (identity-relation set)
  (map
    (lambda (elt) (list elt elt))
    set))

(define compose
  (lambda (rel1 rel2)
    (union*
      (map
        (lambda (pair1)
          (filter
            (lambda (pair2)
              (eqv? (cadr pair1) (car pair2)))
            (lambda (pair2) (list (car pair1) (cadr pair2)))
            rel2))
        rel1))))

;; compose a non-empty bunch of relations:

(define compose-many
  (lambda relations
    (let compose-many ((relations relations))
      (cond
        ((null? relations) (error 'compose-many "too few relations"))
        ((null? (cdr relations)) (car relations))
        (else
          (compose (car relations) (compose-many (cdr relations))))))))

(define (reflexive-transitive-closure reln domain)
  (let loop ((guess (union reln (identity-relation domain))))
    (let 
      ;; use repeated squaring to converge quickly:
      ((next-guess (compose guess guess)))
      ;; since identity \subset guess, we have guess \subset
      ;; next-guess, so the length test is enough -- clever!
      ;; Thanks Ken, I wouldn't have thought of that.
      (if (equal? (length next-guess) (length guess))
        next-guess
        (loop next-guess)))))

;;;;;;;;;; Now let's apply this to the structure at hand. ;;;;;;;;;

(define classnames (map car classes))    ;; assumes no dups

(define edgenames
  (union*
    (map                                ; don't need a filter here
      (lambda (class-spec)
        (map field-name (class-fields class-spec)))
      classes)))

(define get-relation
  (lambda (name)
    (filter
      (lambda (x) (eqv? (car x) name))
      (lambda (x) (list (cadr x) (caddr x)))
      relations)))

(define (edge-relation edgename)
  (filter
    (lambda (x) (and
                  (eqv? (car x) 'field)
                  (eqv? (cadddr x) edgename)))
    (lambda (x) (list (cadr x) (caddr x)))
    relations))

;; these next few guys could as easily be constants, like edgenames.
;; Doing so would simplify first-fields.

(define (C)
  (filter
    (lambda (x) (eqv? (car x) 'field))
    (lambda (x) (list (cadr x) (caddr x)))
    relations))

(define (le)
  (reflexive-transitive-closure
    (get-relation 'subclass)
    classnames))

(define (ge)
  (reflexive-transitive-closure
    (get-relation 'superclass)
    classnames))

;;; and now for the piece de resistance.  We preprocess all the pieces
;;; that are independent of c1 and c2.  These are an association list
;;; associating edges e with their <=e=> relations, and the relation
;;; <=C=>. 

(define first-fields
  ;; do a bunch of computing before you see the arguments:
  (let* ((le (le))
         (ge (ge))
         (C (C))
         (c-table ;; (<=.C.=>)*
           (reflexive-transitive-closure
             (compose-many le C ge) classnames))
         (e-table  ;; association list of e and <=.e.=>.(<=C=>)*
          (map
            (lambda (edgename)
              (list
                edgename
                (compose-many le (edge-relation edgename) ge c-table)))
            edgenames)))
    ;; finished preprocessing.  Now do a comprehension over the edges:
    (lambda (c1 c2)
      (filter
        (lambda (edgename) (is-in? c1 c2 (cadr (assq edgename e-table))))
        (lambda (edgename) edgename)
        edgenames))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; tiny test harness
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; stripped down from the one in EOPL2.

;;;;;; datatype for test items
;; a test consists of a name, a program, and the correct outcome

;; representation: (name program outcome)

(define make-test
  (lambda (name program outcomes)
    (list name  program outcomes)))

(define test->name
  (lambda (l) (list-ref l 0)))

(define test->program
  (lambda (l) (list-ref l 1)))

(define test->outcome
  (lambda (l) (list-ref l 2)))

(define assq-test-list assq)            ; interface procedure for
                                        ; searching 

;;;;;; interface to the test suite database

(define all-tests '())

(define initialize-test-suite!
  (lambda ()
    (set! all-tests '())))

;; replace test of the same name, otherwise add to end:

(define add-to-all-tests!
  (lambda (the-test)
    (let ((name (test->name the-test)))
      (set! all-tests
        (let loop ((tests all-tests))
          (cond 
            ((null? tests)
             (list the-test)) 
            ((eqv? (test->name (car tests)) name)
             (cons the-test (cdr tests)))
            (else (cons (car tests) (loop (cdr tests))))))))))

;;; main entries for adding test items:

(define add-test!
   (lambda (name program correct-answer)
     (add-to-all-tests!
       (make-test name program correct-answer))))

;;; ;;; running tests

;; safely run a test:  run (test pgm); if error is reported,
;; then error-val is returned instead.  That way, if one test bombs,
;; we can keep going.

(define safely-run-experiment-on-program
  (lambda (experiment pgm error-val)
    (call-with-current-continuation
      (lambda (exit)
        ;; redefine error for the
        ;; duration of the check, and reset it
        ;; no matter what.
        (let ((alpha (lambda () (exit error-val))))
          (let ((swap (lambda ()
                        (let ((temp error))
                          (set! error alpha)
                          (set! alpha temp)))))
            (dynamic-wind
              swap
              (lambda () (experiment pgm))
              swap)))))))

(define stop-after-first? #f)
(define show-correct-answers? #t)

(define run-tests
  (lambda ()
    (let ((bugs '()))
      (for-each 
        (lambda (test)
          (let ((name (test->name test))
                (pgm  (test->program test))
                (correct-outcome (test->outcome test)))
            (begin
              (eopl:printf "test: ~a~%~a~%" name pgm)
              (if (eqv? correct-outcome 'dontrun)
                (eopl:printf "test skipped~%~%")
                (let ((ans
                        (safely-run-experiment-on-program
                          eval pgm 'error)))
                  (if show-correct-answers?
                    (eopl:printf "correct outcome: ~a~%" correct-outcome))
                  (eopl:printf "actual outcome:  ")
                  (eopl:printf "~a~%" ans)
                  (if (equal-outcomes? ans correct-outcome)
                    (eopl:printf "correct~%~%")
                    (begin
                      (eopl:printf "incorrect~%~%")
                      ;; stop on first error if stop-after-first? is set:
                      (if stop-after-first? (eopl:error-stop)) 
                      (set! bugs
                        (cons name bugs)))))))))
        all-tests)
      (if (null? bugs)
        (eopl:printf "no bugs found~%")
        (eopl:printf "incorrect answers on tests: ~a~%" bugs)))))


;; this may change from one application to another.  For the moment,
;; we'll make it equal.  If different implementations give answers in
;; different orders, then we'll have to make it do set equality.

(define equal-outcomes? equal?)

;; if silk doesn't have printf, here's a copy:

(define eopl:printf
  (lambda (format . args)
    (let ((len (string-length format)))
      (let loop ((i 0) (args args))
        (let ((output
                (lambda (fn)
                  (fn (car args))
                  (loop  (+ i 2) (cdr args))))
              (outputc
                (lambda (fn)
                  (fn)
                  (loop (+ i 2) args))))
          (if (>= i len) #t
            (let ((c (string-ref format i)))
              (if (char=? c #\~)
                (case (string-ref format (+ i 1))
                  ((#\s) (output write))
                  ((#\a) (output display))
                  ((#\c) (output write-char))
                  ((#\% #\n) (outputc newline))
                  ((#\~) (outputc (lambda () (write-char #\~))))
                  (else
                    (write
                      "error in eopl:printf: unknown format character ")
                    (write-char  (string-ref format (+ i 1)))
                    (write-char #\newline)
                    (eopl:error-stop)))
                (begin
                  (display c)
                  (loop (+ i 1) args))))))))))


;;; end test harness
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; At last, the tests:

(add-test! 'test1 '(first-fields 'library 'magazine) '(publications))

(add-test! 'test2
  '(first-fields 'library 'string)
  '(name publications))

(add-test! 'test3
  '(first-fields 'library 'date)
  '(publications))
