#|<--                                92 columns                                         ->|#


(let ((time-stamp "Time-stamp: <2002-06-11 12:59:38 gregor>"))
  (display (string-append "weaver=microajd-rt.scm: " (substring time-stamp 13 29)))
  (newline))

;;; This is micro-ajd, wihch is an excerpt of ajd.
;;;
;;;	JPs are method calls, having method name, args,
;;;	a link to stack-previous JP, and thunk to continue.
;;;	
;;;	Means of grouping is call and cflow PCDs, with args mechanism.
;;;
;;;	Means of effecting is before advice.

;;;
;;;
;;; A runtime weaving implementation of the aspectj dynamic join-point-model style of
;;; aop.  
;;;
;;; It builds on the language defined in base.  
;;;
;;; The join points are: <n dynamic join point kinds>
;;;
;;; The means of identifying join points is: pointcuts
;;;
;;; The means of affecting behavior at the join points is: before, after and around
;;; advice.
;;;
;;;
;;; Numerous features of AspectJ are missing, including:
;;;
;;;    dispatch on static type in pointcuts
;;;    issingleton, perthis, pertarget, percflow
;;;

(define weave
  (lambda (pgm) ; -> computation
    (fluid-let ((*jp*             (make-bottom-jp))

		(call-method
		 (lambda (method-name obj args)
		   (fluid-let ((*jp* 
				(make-jp 'call method-name *jp* args
					 (lambda (args)
					   (execute-method
					    (lookup-method method-name (object-cname obj)) obj args)))))
		     (execute-matching-advice *jp*
		      (lookup-matching-advice *jp*))))))

      (original-eval-program pgm))))

(define *jp* #f)

(define make-bottom-jp
  (lambda ()
    (make-jp '<invalid> '<invalid> '<invalid> '<invalid> '<invalid>)))

;;; ----------------------------------------------------------------------------
;;; Parsing and loading (eopl calls loading elaboration)
;;;
;;;


(define *advice-decls*  '())

(set! parse-member-decl
  (around-wrap parse-member-decl
    (lambda (original-parse-member-decl enclosing-cname sexp)
      (or (maybe-parse-advice   enclosing-cname sexp)
          (maybe-parse-pointcut enclosing-cname sexp)
          (original-parse-member-decl enclosing-cname sexp)))))

(define maybe-parse-advice
  (lambda (enclosing-cname sexp)
    (and (list? sexp)
         (memq (car sexp) '(before after around))
         (make-advice-decl enclosing-cname (car sexp) (parse-pcd (cadr sexp))
           (parse-body (cddr sexp))))))

(define maybe-parse-pointcut
  (lambda (enclosing-cname sexp)
    (and (list? sexp)
         (memq (car sexp) '(pointcut))
         (make-named-pcd-decl enclosing-cname (cadr sexp) (parse-params (caddr sexp))
	   (parse-pcd (cadddr sexp))))))

(define original-eval-program eval-program)
(define eval-program weave)		; tester's entry

(set! elaborate-class-decl!
  (around-wrap elaborate-class-decl!
    (lambda (original-elaborate-class-decl! decl)
      (let ((decls (class-decl-decls decl)))
	(set! *advice-decls* (append *advice-decls* (collect-if advice-decl?    decls)))
	(original-elaborate-class-decl! decl)))))



;;; ----------------------------------------------------------------------------
;;; Join Points
;;;
;;;

(define-struct jp (kind name stack-previous args thunk))

(define call-jp?       (lambda (x) (jp-kind-test? x 'call)))
(define jp-kind-test?  (lambda (x kind) (and (jp? x) (eq? (jp-kind x) kind))))


;;; ----------------------------------------------------------------------------
;;; Join point designation (Point Cut Designators)
;;;
;;;

(define pointcut-matches?
  (lambda (pcd jp)
    (cond ((call-pcd? pcd)
	   (and (call-jp? jp)
		(name-match? (call-pcd-mname pcd) (jp-name jp))
		(make-matching-pointcut '() '())))
	  
	  ((args-pcd? pcd)
	   (make-matching-pointcut (map param-pname (args-pcd-params pcd))
				   (jp-args jp)))
	  
	  ((cflow-pcd? pcd)
	   (and (valid? jp)
		(or (pointcut-matches? (cflow-pcd-pcd pcd) jp)
		    (pointcut-matches? pcd (jp-stack-previous jp)))))

	  ((and-pcd? pcd)
	   (let loop ((pcds (and-pcd-pcds pcd))
		      (ids  '())
		      (vals '()))
	     (if (null? pcds)
		 (make-matching-pointcut ids vals)
		 (let ((m (pointcut-matches? (car pcds) jp)))
		   (and (matching-pointcut? m)
			(loop (cdr pcds)
			      (append ids (matching-pointcut-ids  m))
			      (append vals (matching-pointcut-vals m)))))))))))

(define-struct matching-pointcut (ids vals))

(define name-match?
  (lambda (pattern name)
    (eq? pattern name)))

(define valid?
  (lambda (val)
    (not (eq? val '<invalid>))))

;;;
;;; this is the weaver
;;;
;;; note that in AspectJ, the advice rules are somewhat more complex.  
;;; Two important differences are that:
;;;
;;;    - advice is sorted by precedence
;;;
;;;    - each proceed causes a new advice lookup
;;;

(define execute-matching-advice
  (lambda (tjp madvice)
    (let ((args    (jp-args tjp))
	  (thunk   (jp-thunk tjp)))
      (execute-befores madvice tjp args)
      (thunk args))))

(define execute-befores
  (lambda (m-advs tjp args)
    (for-each (lambda (m-adv)
                (eval-body (advice-decl-body (matching-advice-adv-decl m-adv))
			   (make-advice-env m-adv tjp args)))
              m-advs)))

(define make-advice-env
  (lambda (m-adv tjp args)
    (let ((ids  (matching-advice-id m-adv))
	  (vals (matching-advice-vals m-adv)))
      (new-env (cons 'tjp ids) (cons tjp vals)))))

(define lookup-matching-advice
  (lambda (jp)
    (filter-map (lambda (adv)
		  (let ((m (pointcut-matches? (advice-decl-pcd adv) jp)))
		    (and (matching-pointcut? m)
			 (make-matching-advice adv
					       (matching-pointcut-ids m)
					       (matching-pointcut-vals m)))))
		*advice-decls*)))


(define-struct matching-advice (adv-decl id vals))

;;;
;;; The next 11 definitions (3 variable definitions and 8 wrappers) are, in
;;; effect, an aspect that makes join points and calls the weaver.  (I would
;;; have liked to put them before the preceding weaver code, but I can't do
;;; that with this wrapper implementation.)
;;;
;;; Note that these three dynamic variables are ONLY used within these 11
;;; definitions.  They are like a "lexically extent dynamically scoped" variable
;;; of these wrappers.
;;;
;;; Whatever you do, don't get the idea of not passing tjp to push-jp, thinking
;;; it could just get it out of *jp*  Consider what would happen if we removed
;;; cflow like features from this language.  *jp* would go away, but we would still
;;; need to be able to call push-jp.
;;;
;
;(aspect weaver-calling
;

; (set! call-method
;   (around-wrap call-method
;     (lambda (original-call-method mname obj args)
;       (fluid-let ((*jp* 
; 		   (make-jp 'call mname *jp* args
; 			    (lambda (args) ;thunk
; 			      (original-call-method mname obj args)))))
; 	(push-jp *jp*)))))


;;;
;;; make-jp-with-context-and-dispatch is supposed to make the join point and then call
;;; dispatch-advice.  In its full-blown version, it handles the context information of
;;; 
;;;   jp-within
;;;   jp-stack-previous
;;;   jp-time-previous
;;;
;;; But a simpler version of it could just look like:
;;;
;;;   (define make-jp-with-context-and-dispatch
;;;     (lambda (kind name this target args thunk)
;;;       (push-jp (make-jp kind name this target args thunk))))
;;;
;;;
(define *jp*  #f)

;;;
;;; this just makes sure there is an outermost (lowermost whatever) join point
;;;
; (set! eval-program
;   (around-wrap eval-program
;     (lambda (original-eval-program pgm)
;       (fluid-let ((*jp*
; 		    (make-jp '<invalid> '<invalid>
; 			     '<invalid> '<invalid> '<invalid>))) ;this is NEVER tjp
; 	(original-eval-program pgm)))))


;) ;end of weaver-calling aspect
