#|<--                               90 columns                                        ->|#

(let ((time-stamp "Time-stamp: <2002-06-11 12:59:46 gregor>"))
  (display (string-append "weaver=trav-rt.scm: " (substring time-stamp 13 29)))
  (newline))

;;; A: class graph
;;;    ID: class and field names
;;;    effect: return field values
;;;
;;; B: traversal specification
;;;    ID: from-to specification, (visitor, and customized class graph)
;;;    effect: navigate traversal, (run visitor methods)
;;;
;;; X: traversal over object graphs
;;;    JPs: nodes in traversal path (i.e., objects)
;;;

(define trav:weave
  (lambda (trav-spec root visitor); -> traversal
    (let arrive ((obj  root)                                     ;arrival at obj is a JP
                 (path (make-path (object-cname root))))
      (call-visitor visitor obj)                                 ;assumes root is OK
      (for-each (lambda (fname)
                  (let* ((next-obj   (get-field fname obj))
                         (next-cname (object-cname next-obj))
                         (next-path  (extend-path path next-cname)))
                    (if (match? next-path trav-spec)
                        (arrive next-obj next-path))))           ; EFF: continue
                (object->fnames obj)))))

(define match?
  (lambda (path trav-spec)
    (let ((residual-spec
           (match-path path trav-spec)))
      (match-class-graph? residual-spec
                      (path-last-cname path)))))

(define match-path			;matches a path to traversal specification, and
  (lambda (path trav-spec)		;retrunes a specification yet to be traversed
    (let loop ((cnames    (path->cnames path))
	       (trav-spec trav-spec))
      (if (null? cnames) trav-spec
	  (loop (cdr cnames)
		(match-cname (car cnames) trav-spec))))
    ;; the above 5 lines are equivalent to:
    ;; (fold match-cname (path->cnames path) trav-spec)
    ))

(define match-cname
  (lambda (cname trav-spec)
    (cond ((null? trav-spec) #f)
          ((from-spec? trav-spec)
           (if (subclass? cname (from-spec-cname trav-spec))
               (spec-next trav-spec)
               (error 'match "path ~s doesn't match spec ~s."
                      cname trav-spec)))
          ((to-spec? trav-spec)
           (if (subclass? cname (to-spec-cname trav-spec))
               (make-or-spec '()
			     (list trav-spec (spec-next trav-spec)))
               trav-spec))
          ((or-spec? trav-spec)
           (make-or-spec '()
	    (map (lambda (spec)
		   (and spec
			(match-cname cname
				     (append spec (spec-next trav-spec)))))
		 (or-spec-specs trav-spec)))))))

(define match-class-graph?		;possibility to make a matching path
  (lambda (trav-spec root-cname)	;in terms of class graph
    (let loop ((trav-spec trav-spec)
               (cname root-cname))
      (cond ((null? trav-spec) #t)
            ((eq? trav-spec #f) #f)
            ((to-spec? trav-spec)
             (and (reachable? cname
			      (to-spec-cname trav-spec))
                  (loop (spec-next trav-spec)
                        (to-spec-cname trav-spec))))
            ((or-spec? trav-spec)
             (some? (lambda (spec) (loop spec cname))
                    (or-spec-specs trav-spec)))))))

(define reachable?
  (lambda (from-cname to-cname)
    (some? (lambda (reachable-cname)
	     (subclass? reachable-cname to-cname))
	   (reachable-cnames from-cname))))

(define reachable-cnames		;a set of reachable classes from C, say R(C), is
  (lambda (from-cname)			;a minimal set s.t. (1) C is in R(C), (2) 
    (transitive-closure			;for any C' in R(C), all subclasses of C' are
     (lambda (cname)			;in R(C), and (3) for any C' in R(C), all
       (apply set-join			;direct reachable classes are in R(C).
	      (map direct-reachable
		   (cname->subclass-cnames cname))))
     from-cname)))

(define direct-reachable		      ;return a list of types of instance 
  (lambda (cname)			      ;variables (including inherited ones)
    (remove-duplicates			      ;in the class.  this takes information
     (map field-type (cname->fields cname)))));out from *classes* global variable

(define cname->subclass-cnames
  (lambda (cname)
    (collect-if (lambda (maybe-subclass-cname)
		  (memq cname (cname-ancestors maybe-subclass-cname)))
		(map class-cname *classes*))))    

(define cname-ancestors
  (lambda (cname)
    (if cname
	(cons cname 
	      (cname-ancestors (class-sname (lookup-class cname))))
	'())))

(define transitive-closure
  (lambda (get-neighbors root)
    (let loop ((unchecked (get-neighbors root))
	       (checked   '()))
      (if (null? unchecked)
	  checked
	  (let* ((neighbors     (get-neighbors (car unchecked)))
		 (new-unchecked (set-difference
				 (set-difference neighbors unchecked)
				 checked)))
	    (loop (append new-unchecked (cdr unchecked))
		  (cons (car unchecked) checked)))))))

;; a path is implemented as a reversed list of class names
(define make-path       (lambda (root-cname) (list root-cname)))
(define path-last-cname car)
(define extend-path     (lambda (path cname) (cons cname path)))
(define path->cnames    (lambda (path) (reverse path)))

;;; traversal spac parser
;;; SPEC ::= (from    NAME            ,@SPEC) 
;;;        | (to      NAME            ,@SPEC)
;;;        | (through NAME            ,@SPEC)
;;;        | (or      (SPEC SPEC ...) ,@SPEC)

; (define-struct spec (next))
; (define-struct (from-spec struct:spec) (cname))
; (define-struct (to-spec   struct:spec) (cname))
; (define-struct (or-spec   struct:spec) (specs))

(define from-spec?
  (lambda (spec) (and (pair? spec) (eq? (car spec) 'from))))
(define to-spec?
  (lambda (spec) (and (pair? spec) (memq (car spec) '(to through)))))
(define or-spec?
  (lambda (spec) (and (pair? spec) (eq? (car spec) 'or))))
(define spec-next          cddr)
(define from-spec-cname    cadr)
(define to-spec-cname      cadr)
(define or-spec-specs      cadr)
(define make-or-spec (lambda (next specs) `(or ,specs ,@next)))

;;;
  
(define call-visitor			        ;visit methods have to discriminate 
  (lambda (visitor obj)			        ;object types for accurate behaviors
    (call-method 'visit visitor (list obj))))

(set! object-cname			        ;give cnames for non-object values
  (around-wrap object-cname
   (lambda (original-object-cname object)
     (cond ((object?  object) (original-object-cname object))
	   ((integer? object) 'int)
	   ((string?  object) 'string)
	   (else	      'object)))))

(define object->fnames			
  (lambda (object)
    (cname->fnames (object-cname object))))

(define get-field
  (lambda (fname object)
    (get-field-value fname (object-cname object) object)))

;;; add the traverse primitive

(set! *primitives*
      `((traverse ,(lambda args (apply trav:weave args)))
	,@*primitives*))

